﻿using FreeRedis;

namespace DBHelper.Redis
{
    /// <summary>
    /// Redis
    /// </summary>
    public static class RedisHub
    {
        /// <summary>
        /// 连接集合
        /// </summary>
        private static List<RedisClientList> RedisClients = new List<RedisClientList>();

        /// <summary>
        /// 默认上下文类类型
        /// </summary>
        public static string DefaultContextType { get; private set; } = null!;

        /// <summary>
        /// 默认监控上下文类类型
        /// </summary>
        public static string MonitorContextType { get; private set; } = null!;

        /// <summary>
        /// 默认调试上下文类类型
        /// </summary>
        public static string DebugContextType { get; private set; } = null!;

        /// <summary>
        /// 设置默认链接
        /// </summary>
        /// <param name="key"></param>
        public static void SetDefaultContextType(string key)
        {
            DefaultContextType = key;
        }

        /// <summary>
        /// 设置监控链接
        /// </summary>
        /// <param name="key"></param>
        public static void SetMonitorContextType(string key)
        {
            MonitorContextType = key;
        }

        /// <summary>
        /// 设置监控链接
        /// </summary>
        /// <param name="key"></param>
        public static void SetDebugContextType(string key)
        {
            DebugContextType = key;
        }

        /// <summary>
        /// 默认链接
        /// </summary>
        public static RedisClient Default
        {
            get
            {
                {
                    if (DefaultContextType == null)
                        throw new Exception("请先设置默认RedisDB库,调用静态方法SetDefaultDbContextType()");
                    return Context(DefaultContextType);
                }
            }
        }

        /// <summary>
        /// 监控连接
        /// </summary>
        public static RedisClient Monitor
        {
            get
            {
                {
                    if (MonitorContextType == null)
                        throw new Exception("请先设置监控RedisDB库,调用静态方法SetMonitorContextType()");
                    return Context(MonitorContextType);
                }
            }
        }

        /// <summary>
        /// 监控连接
        /// </summary>
        public static RedisClient Debug
        {
            get
            {
                {
                    if (MonitorContextType == null)
                        throw new Exception("请先设置DebugRedisDB库,调用静态方法SetDebugContextType()");
                    return Context(MonitorContextType);
                }
            }
        }

        /// <summary>
        /// 创建一个连接
        /// </summary>
        /// <param name="connectionString"> 连接字符串</param>
        /// <param name="key">标识</param>
        /// <param name="defaultClient"> 是否为默认连接，默认为false</param>
        /// <returns></returns>
        public static RedisClient CreateContext(string connectionString, string key, bool defaultClient = false)
        {
            var ctx = RedisClients.FirstOrDefault(v => v.Key == key);
            if (ctx != null) return ctx.Client;
            ctx = new RedisClientList(key, new RedisClient(connectionString), connectionString);
            RedisClients.Add(ctx);
            if (defaultClient)
            {
                SetDefaultContextType(key);
            }
            return ctx.Client;
        }

        /// <summary>
        /// 获取指定的连接
        /// </summary>
        /// <param name="key"></param>
        /// <returns></returns>
        /// <exception cref="Exception"></exception>
        public static RedisClient Context(string key)
        {
            var ctx = RedisClients.FirstOrDefault(v => v.Key == key);
            if (ctx == null) throw new Exception("没有对应的链接，请先调用创建");
            return ctx.Client;
        }

        /// <summary>
        /// 检查Redis中是否有对应key且value不为空
        /// </summary>
        /// <param name="redisClient">redis连接</param>
        /// <param name="key">要检查的Key</param>
        /// <returns>
        /// 1.key不存在,创建这个key value默认为三个空格符。返回null
        /// 2.key存在单value为空 返回null
        /// 3.key存在value不为空 返回获取到的值
        /// </returns>
        public static string? Check(this RedisClient redisClient, string key)
        {
            var result = redisClient.Get(key);
            if (!string.IsNullOrEmpty(result)) return result;
            redisClient.Set(key, "   ");
            return null;
        }
    }
}
